/*
 * Copyright (C) 2010 Andrew Kozin AKA Stanson <stanson@pangolin.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * CarLock engine
 *
 *                       Authentication sequence:
 *
 *            Application                            Device
 *
 *                                  <--          OK: Hello,state,challenge
 * PASS sha1( pass + challenge )    -->
 *                                  <--          OK: Access granted or ERR: Access denied
 */

#include <carlock.rsg>
#include <coemain.h>
#include <es_enum.h>
#include <commdbconnpref.h>
#include <commdb.h>
#include <hash.h>
#include <utf.h>

#include "carlock_appui.h"
#include "carlock_engine.h"
#include "carlock_appview.h"
#include "carlock_console.h"
#include "carlock_settings.h"

_LIT8( KCRLF, "\n" );
_LIT8( KREPLY_OK, "OK" );
_LIT8( KREPLY_OPENED, "Opened");
_LIT8( KREPLY_CLOSED, "Closed");
_LIT8( KREPLY_HELLO, "Hello");

CCarLockEngine* CCarLockEngine::NewL( CCarLockAppUi* parent )
{
    CCarLockEngine* self = new ( ELeave ) CCarLockEngine( parent );

    CleanupStack::PushL( self );
    self->ConstructL();
    CleanupStack::Pop( self );

    return self;
}

CCarLockEngine::CCarLockEngine( CCarLockAppUi* parent ) : 
    CActive( CActive::EPriorityStandard )
{
    iParent   = parent;
    iReconn   = 0;
    iState    = EDisconnected;
    iCmdSent  = ECmdNone;

    iDevState = EDevUnknown;

    iBuffer.Zero();

    CActiveScheduler::Add( this );
}

void CCarLockEngine::ConstructL()
{
    iToneUtility = CMdaAudioToneUtility::NewL( *this );
    iTimer = CCarLockTimer::NewL( *this );
    User::LeaveIfError( iSocketServer.Connect() );
}

CCarLockEngine::~CCarLockEngine()
{
    iSocket.CancelAll();

    Cancel();

    iSocket.Close();
    iSocketServer.Close();

    delete iTimer;
}

void CCarLockEngine::DoCancel() { }

void CCarLockEngine::MatoPrepareComplete( TInt /*aError*/ )
{
    iToneUtility->SetVolume( iToneUtility->MaxVolume() );
    iToneUtility->Play();
}

void CCarLockEngine::MatoPlayComplete(TInt /*aError*/) { }

void CCarLockEngine::RunL()
{
    iTimer->Cancel();

    switch ( iState )
    {
	case EResolving: // DNS look up finished

	    if( iStatus != KErrNone )
	    {
		iParent->iConsole->Print( _L("Device %S not found, %d"), &iDevice, iStatus.Int() );
		iResolver.Close();
		Disconnect( ETrue );
		break;
	    }
	    
	    iParent->iConsole->Print( _L("Found %S"), &iNameEntry().iName );
	    
	    if( iNameEntry().iName != iDevice )
	    {
		iResolver.Next( iNameEntry, iStatus );
		SetActive();
		break;
	    }

	    // Device found
	    iResolver.Close();

	    iDevAddr.Des() = static_cast<TBTSockAddr>(iNameEntry().iAddr).BTAddr().Des();
	    SocketConnect( iDevAddr, iChannel );

	    break;

	case EConnecting:
	    if( iStatus != KErrNone )
	    {
		switch( iStatus.Int() )
		{
		    case -6004: /* EPageTimedOut */
			iParent->iConsole->Print( _L("ERR: Page scan timeout") );
			break;
		    default:
			iParent->iConsole->Print( _L("Connect error %d"), iStatus.Int() );
			break;
		}
		Disconnect( ETrue );
		break;
	    }

	    iToneUtility->PrepareToPlayTone( 1320, 50000 );
	    iParent->iAppView->SetDevice( iAddress );
	    SetState( EConnected );

	    iBuffer.Zero();
	    RequestData();
	    
	    break;
              
	case EConnected:
	    if( iStatus != KErrNone )
	    {
		switch( iStatus.Int() )
		{
		    case KErrTimedOut:
			Stat();
			return;
		    case KErrHCILinkDisconnection:
			iParent->iConsole->Print( _L("ERR: HCI link disconnection") );
			break;
		    default:
			iParent->iConsole->Print( _L("Rx error %d"), iStatus.Int() );
			break;
		}
		Disconnect( ETrue );
		break;
	    }

	    iParent->iAppView->FlashLed();

	    {
		TInt i;

		for( i = 0; i < iRcvdLen(); i++ )
		{
		    if( iBuffer[i] && iBuffer[i] != '\n' && iBuffer[i] != '\r' )
		    {
			iReply.Append( iBuffer[i] );
		    }
		    else
		    { /* we got whole reply */
			if( iReply.Length() > 0 )
			{
			    HandleReply( iReply );
			    return;
			}
		    }
		}
	    }

	    iBuffer.Zero();
	    RequestData();
	    break;

	case ESending:
	    if( iStatus != KErrNone )
	    {
		iParent->iConsole->Print( _L("Tx error %d"), iStatus.Int() );
		Disconnect( ETrue );
		break;
	    }

	    SetState( EConnected );
	    RequestData();
    	    break;

	case EDisconnected:
	    if( iStatus == KErrTimedOut )
	    {
		SocketConnect( iDevAddr, iChannel );
	    }
	    break;

    	default:
	    break;
    }
}

/* Time-out timer callback */
void CCarLockEngine::TimerExpired()
{
    TRequestStatus* status = &iStatus;
    if( iState == EConnected ) iSocket.CancelRead();
    Cancel();
    SetActive();
    User::RequestComplete( status, KErrTimedOut );
}

void CCarLockEngine::HandleReply( TDes8& aReply )
{
    CnvUtfConverter::ConvertToUnicodeFromUtf8( iUnicode, aReply );
    iParent->iConsole->Print( _L("> %S"), &iUnicode );

    if( aReply.Mid( 0, 2 ) == KREPLY_OK )
    {
	if( aReply.Find( KREPLY_OPENED ) != KErrNotFound )
	{
	    iDevState = EDevOpened;
	    iParent->iAppView->SetStat( _L("open") );
	}
	else if( aReply.Find( KREPLY_CLOSED ) != KErrNotFound )
	{
	    iDevState = EDevClosed;
	    iParent->iAppView->SetStat( _L("closed") );
	}

	if( aReply.Find( KREPLY_HELLO ) != KErrNotFound )
	{
	    TBuf8<16> challenge;
	    TInt pos = aReply.LocateReverse(',');
	    if( pos != KErrNotFound )
	    {
		challenge.Copy( aReply.Mid( pos + 1, 8 ) );
		SendPass(iParent->iSettings->GetPassword(), challenge);
		aReply.Zero();
		return;
	    }
	}

	switch( iCmdSent )
	{
	    case ECmdStat: /* f.e.: OK: <DevState>,<VRef>,<Vdd> */
		{
		    TLex8 lex;
		    TReal32 Vref, Vdd;
		    TPtrC8 ptr;
		    TInt pos;
		    
		    pos = aReply.Locate( ',' );
		    if( pos == KErrNotFound ) break;
		    ptr.Set( aReply.Mid( pos + 1 ) );
		    lex = ptr;
		    lex.Val( Vref );

		    pos = ptr.Locate( ',' );
		    if( pos == KErrNotFound ) break;
		    ptr.Set( ptr.Mid( pos + 1 ) );
		    lex = ptr;
		    lex.Val( Vdd );

		    Vref = 1.25 * 255.0 / Vref;
		    Vdd = Vref * Vdd / 255.0;
		    Vdd *= iParent->iSettings->GetVddScale();

		    iParent->iAppView->SetVoltage( Vdd );
		}
		break;
	    default:
		break;
	}
    }

    iCmdSent = ECmdNone;
    
    iTimer->After( iParent->iSettings->GetSearchInt() * 1000 );

    aReply.Zero();
    RequestData();
}

void CCarLockEngine::Disconnect( TBool aRetry )
{
    iCmdSent = ECmdNone;

    switch( iState )
    {
	case EConnected:
	case ESending:
	    iSocket.CancelAll();
	    Cancel();
	    iToneUtility->PrepareToPlayTone( 880, 50000 );
	    iParent->iAppView->SetDevice( _L("---") );
	    iParent->iAppView->SetStat( _L("unknown") );
	    iParent->iConsole->Print( _L("Disconnected") );

	case EConnecting:
	    iSocket.Close();

	case EResolving:
	    SetState( EDisconnected );

	case EDisconnected:
	    if( aRetry ) iTimer->After( iParent->iSettings->GetSearchInt() * 1000 );
	    break;

	default:    
	    break;
    }
}

TInt CCarLockEngine::Connect( const TDesC& aDevice, const TDesC& aAddress, TInt aChannel )
{
    TInt res;
    TProtocolDesc pdesc;

    if( State() != EDisconnected || IsActive() ) return KErrInUse;

    iDevice = aDevice;
    iChannel = aChannel;

    res = iDevAddr.SetReadable( aAddress );
    if( res > 0 )
    {

	SocketConnect( iDevAddr, iChannel );

	return KErrNone;
    }

    /* bad address, try to search device */

    iSocketServer.FindProtocol( _L("BTLinkManager"), pdesc );

    res = iResolver.Open( iSocketServer, pdesc.iAddrFamily, pdesc.iProtocol );

    if( res != KErrNone )
    { // resolver failed
	iParent->iConsole->Print( _L("Resolver open failed") );
	return res;
    }

    iSockAddr.SetAction( KHostResInquiry | KHostResName | KHostResIgnoreCache );

    iParent->iConsole->Print( _L("Searching %S..."), &iDevice );
    SetState( EResolving );

    iResolver.GetByAddress( iSockAddr, iNameEntry, iStatus );
    SetActive();

    return KErrNone;
}

void CCarLockEngine::SocketConnect( TBTDevAddr& aDevAddr, TInt aChannel )
{
    TInt res;

    aDevAddr.GetReadable( iAddress );
    iParent->iConsole->Print( _L("Connecting to %S"), &iAddress );
    SetState( EConnecting );

    res = iSocket.Open( iSocketServer, _L("RFCOMM") );

    if( res != KErrNone )
    {
        iParent->iConsole->Print( _L("Socket open failed") );
	TRequestStatus* status = &iStatus;		
	Cancel();
	SetActive();
	User::RequestComplete( status, res );
        return;
    }

    TBTSockAddr addr;
    TBTServiceSecurity sec;

    sec.SetEncryption( ETrue );

    addr.SetBTAddr( aDevAddr );
    addr.SetPort( aChannel );
    addr.SetSecurity( sec );
				
    iSocket.Connect( addr, iStatus );
    SetActive();
}

TCarLockState CCarLockEngine::State()
{
    return iState;
}

void CCarLockEngine::SetState( TCarLockState aState )
{
    iState = aState;
}

void CCarLockEngine::RequestData()
{
    if( State() != EConnected ) return;
    iSocket.RecvOneOrMore( iBuffer, 0, iStatus, iRcvdLen );
    SetActive();
}

/* Send formatted string to device */

void CCarLockEngine::Send( TRefByValue<const TDesC8> aFmt, ... )
{
    VA_LIST list;

    if( State() != EConnected ) return;

    iMessage.Zero();

    VA_START( list, aFmt );
    iMessage.AppendFormatList( aFmt, list );
    VA_END( list );

    CnvUtfConverter::ConvertToUnicodeFromUtf8( iUnicode, iMessage );
    iParent->iConsole->Print( _L("< %S"), &iUnicode );

    iMessage.Append( KCRLF );

    iTimer->Cancel();

    iSocket.CancelRead();

    if( IsActive() ) Cancel();

    SetState( ESending );
    
    iSocket.Send( iMessage, 0, iStatus, iSentLen );
    SetActive();
}

void CCarLockEngine::Bin2Hex( TDes8& dst, const TDesC8& src )
{
    int i, len = src.Length();
    
    dst.Zero();
    for( i = 0; i < len; i++ )
    {
	dst.AppendFormat( _L8("%02X"), src[i] );
    }
}

/* Send PASS command */
void CCarLockEngine::SendPass( const TDesC& pass, const TDesC8& challenge )
{
    TBuf8<128>	upass;
    TBuf8<40>	hash;
    CMD5*	md5 = CMD5::NewL();

    CnvUtfConverter::ConvertFromUnicodeToUtf8( upass, pass );
    md5->Update( upass );
    md5->Update( challenge );
    Bin2Hex( hash, md5->Final() );

    Send( _L8("PASS %S"), &hash );

    delete md5;
}

/* Open */
void CCarLockEngine::Open()
{
    iCmdSent = ECmdOpen;
    Send( _L8("OPEN %d"), iParent->iSettings->GetDelay() );
}

/* Close */
void CCarLockEngine::Close()
{
    iCmdSent = ECmdClose;
    Send( _L8("CLOSE %d"), iParent->iSettings->GetDelay() );
}

/* Stat */
void CCarLockEngine::Stat()
{
    iCmdSent = ECmdStat;
    Send( _L8("STAT") );
}

TCarLockDevState CCarLockEngine::DevState()
{
    return iDevState;
}
